package jp.hishidama.swing.undo;

import java.awt.event.*;
import java.util.ArrayList;
import java.util.List;

import javax.swing.event.*;
import javax.swing.text.*;
import javax.swing.undo.*;

/**
 * eLXgUNDO}l[W[.
 * <p>
 * JTextComponentɕ͂UNDO/REDO@\ǉׂ̃}l[W[B
 * </p>
 * <p>
 * WUndoManagerł́A1UNDO/REDOɂȂB<br>
 * UndoManagerł́Ax̌Ăяoœށi܂蕶̒ǉE폜jUNDO/REDOAčsƂɂAPʂłUNDO/
 * REDOB
 * </p>
 *
 * <p>
 * <a target="hishidama" href=
 * "http://www.ne.jp/asahi/hishidama/home/tech/soft/java/swing/UndoManager.html"
 * >gp</a>
 * </p>
 *
 * @author <a target="hishidama"
 *         href="http://www.ne.jp/asahi/hishidama/home/tech/soft/java/swing.html"
 *         >Ђ</a>
 * @since 2007.02.24
 * @version 2009.04.12 UndoManagerInstallǉ
 * @version 2009.04.24 CompoundEditgɕύX
 */
public class TextUndoManager extends UndoManager implements
		UndoableEditListener, UndoManagerInstall {
	private static final long serialVersionUID = 5375924557489349842L;

	protected UndoManagerUtil umu;

	/**
	 * .
	 * <p>
	 * JTextComponentɓUndoManagero^B
	 * </p>
	 *
	 * @param tc
	 *            eLXgR|[lgiJEditorPaneEJTextFieldEJTextAreaj
	 */
	public void install(JTextComponent tc) {
		if (tc == null)
			return;

		if (umu == null) {
			umu = new UndoManagerUtil(this);
		}
		umu.installTo(tc);

		if (handler == null)
			handler = createHandler();
		tc.addKeyListener(handler);
		tc.addMouseListener(handler);
		tc.addFocusListener(handler);
		update();
	}

	/**
	 * I.
	 * <p>
	 * JTextComponent瓖UndoManager폜B
	 * </p>
	 *
	 * @param tc
	 *            eLXgR|[lgiJEditorPaneEJTextFieldEJTextAreaj
	 */
	public void uninstall(JTextComponent tc) {
		if (tc == null)
			return;

		if (umu != null) {
			umu.uninstall(tc);
			// umu = null;
		} else {
			Document doc = tc.getDocument();
			doc.removeUndoableEditListener(this);
		}

		tc.removeKeyListener(handler);
		tc.removeMouseListener(handler);
		tc.removeFocusListener(handler);
	}

	protected EventHandler handler;

	protected EventHandler createHandler() {
		return new EventHandler();
	}

	@Override
	public void undoableEditHappened(UndoableEditEvent e) {
		super.undoableEditHappened(e);
		update();
	}

	/**
	 * UndoeXV.
	 * <p>
	 * Undo/Redoj[ɊւύXsꂽƂɌĂ΂B
	 * </p>
	 * <p>
	 * TuNXŃI[o[ChāAUndo/Redoj[enabled̍XVȂǂsB
	 * </p>
	 */
	public void update() {
	}

	@Override
	public synchronized void discardAllEdits() {
		super.discardAllEdits();
		update();
	}

	/**
	 * @see CompoundEdit
	 */
	static class SameEdit extends AbstractUndoableEdit {
		private static final long serialVersionUID = 3096295069062300124L;

		protected boolean inProgress = true;
		protected List<UndoableEdit> edits = new ArrayList<UndoableEdit>();

		public SameEdit(UndoableEdit e) {
			edits.add(e);
		}

		@Override
		public void undo() throws CannotUndoException {
			super.undo();
			for (int i = edits.size() - 1; i >= 0; i--) {
				UndoableEdit e = edits.get(i);
				e.undo();
			}
		}

		@Override
		public void redo() throws CannotRedoException {
			super.redo();
			for (UndoableEdit e : edits) {
				e.redo();
			}
		}

		protected UndoableEdit lastEdit() {
			int count = edits.size();
			return edits.get(count - 1);
		}

		@Override
		public void die() {
			for (int i = edits.size() - 1; i >= 0; i--) {
				UndoableEdit e = edits.get(i);
				e.die();
			}
			super.die();
		}

		@Override
		public boolean addEdit(UndoableEdit anEdit) {
			if (!inProgress) {
				return false;
			}
			UndoableEdit last = lastEdit();
			if (last.getPresentationName().equals(anEdit.getPresentationName())) {
				edits.add(anEdit);
				return true;
			}
			return false;
		}

		public void end() {
			inProgress = false;
		}

		public boolean isInProgress() {
			return inProgress;
		}

		@Override
		public boolean isSignificant() {
			for (UndoableEdit e : edits) {
				if (e.isSignificant()) {
					return true;
				}
			}
			return false;
		}

		@Override
		public String getPresentationName() {
			UndoableEdit last = lastEdit();
			return last.getPresentationName();
		}

		@Override
		public String getUndoPresentationName() {
			UndoableEdit last = lastEdit();
			return last.getUndoPresentationName();
		}

		@Override
		public String getRedoPresentationName() {
			UndoableEdit last = lastEdit();
			return last.getRedoPresentationName();
		}

		@Override
		public String toString() {
			return super.toString() + " inProgress: " + inProgress + " edits: "
					+ edits;
		}
	}

	/**
	 * UNDOXgɋ؂.
	 */
	public void addSeparator() {
		SameEdit last = (SameEdit) lastEdit();
		if (last != null) {
			last.end();
		}
	}

	@Override
	public synchronized boolean addEdit(UndoableEdit anEdit) {
		UndoableEdit last = lastEdit();
		if (last != null && last.addEdit(anEdit)) {
			return true;
		}
		SameEdit se = new SameEdit(anEdit);
		return super.addEdit(se);
	}

	@Override
	public synchronized void undo() throws CannotUndoException {
		addSeparator();
		if (!canUndo())
			return;

		super.undo();

		update();
	}

	@Override
	public synchronized void redo() throws CannotRedoException {
		addSeparator();
		if (!canRedo())
			return;

		super.redo();

		update();
	}

	protected class EventHandler implements KeyListener, MouseListener,
			FocusListener {

		@Override
		public void keyTyped(KeyEvent e) {
		}

		@Override
		public void keyPressed(KeyEvent e) {
			switch (e.getKeyCode()) {
			case KeyEvent.VK_Z:
				// zKeyPressed(e);
				break;
			case KeyEvent.VK_Y:
				// yKeyPressed(e);
				break;
			case KeyEvent.VK_UP:
			case KeyEvent.VK_DOWN:
			case KeyEvent.VK_LEFT:
			case KeyEvent.VK_RIGHT:
			case KeyEvent.VK_HOME:
			case KeyEvent.VK_END:
			case KeyEvent.VK_PAGE_UP:
			case KeyEvent.VK_PAGE_DOWN:
				addSeparator();
				break;
			}
		}

		/**
		 * ZL[.
		 * <p>
		 * ANZ[^L[UNDOꍇ́A\bhI[o[ChĖɏoB
		 * </p>
		 *
		 * @param e
		 *            L[Cxg
		 */
		protected void zKeyPressed(KeyEvent e) {
			if (e.isControlDown()) {
				undo();
				e.consume();
			}
		}

		/**
		 * YL[.
		 * <p>
		 * ANZ[^L[REDOꍇ́A\bhI[o[ChĖɏoB
		 * </p>
		 *
		 * @param e
		 *            L[Cxg
		 */
		protected void yKeyPressed(KeyEvent e) {
			if (e.isControlDown()) {
				redo();
				e.consume();
			}
		}

		@Override
		public void keyReleased(KeyEvent e) {
		}

		@Override
		public void mouseClicked(MouseEvent e) {
		}

		@Override
		public void mousePressed(MouseEvent e) {
			// {́AeLXg̃J[\̈ʒuςƂ؂肽
			// ł͔fłȂ̂ŁAɋ؂
			addSeparator();
		}

		@Override
		public void mouseReleased(MouseEvent e) {
		}

		@Override
		public void mouseEntered(MouseEvent e) {
		}

		@Override
		public void mouseExited(MouseEvent e) {
		}

		@Override
		public void focusGained(FocusEvent e) {
			addSeparator();
		}

		@Override
		public void focusLost(FocusEvent e) {
			addSeparator();
		}
	}
}
